import datetime
import pathlib
import pytest
from metro_db import SQLiteDB
from enum import IntEnum


class Position(IntEnum):
    CATCHER = 2
    FIRST_BASE = 3
    SECOND_BASE = 4
    THIRD_BASE = 5
    SHORTSTOP = 6
    LEFT_FIELD = 7


@pytest.fixture()
def demo_db():
    path = pathlib.Path('demo.db')
    db = SQLiteDB(path, default_type='int')
    db.tables['batters'] = ['id', 'name', 'year', 'hits', 'position']
    db.field_types['name'] = 'text'
    db.field_types['position'] = 'Position'
    db.register_custom_enum(Position)
    db.update_database_structure()
    for values in [
        ['Olerud', 1998, 197, Position.FIRST_BASE],
        ['Piazza', 1998, 137, Position.CATCHER],
        ['Alfonzo', 1998, 155, Position.THIRD_BASE],
        ['Olerud', 1999, 173, Position.FIRST_BASE],
        ['Piazza', 1999, 162, Position.CATCHER],
        ['Alfonzo', 1999, 191, Position.SECOND_BASE],
        ['Zeile', 2000, 146, Position.FIRST_BASE],
        ['Piazza', 2000, 156, Position.CATCHER],
        ['Alfonzo', 2000, 176, Position.SECOND_BASE],
    ]:
        db.execute('INSERT INTO batters (name, year, hits, position) VALUES(?, ?, ?, ?)', values)

    yield db
    db.dispose()


def test_lookup_all(demo_db):
    values = list(demo_db.lookup_all('name', 'batters', {'year': 1999}))
    assert len(values) == 3
    assert 'Olerud' in values
    assert 'Piazza' in values
    assert 'Alfonzo' in values


def test_lookup(demo_db):
    assert demo_db.lookup('hits', 'batters', {'year': 1999, 'name': 'Alfonzo'}) == 191

    # Contrived examples to test quote handling
    assert demo_db.lookup('hits', 'batters', {'year': 1999, 'name': 'O\'Brien'}) is None
    assert demo_db.lookup('hits', 'batters', {'year': 1999, 'name': '"O\'Brien"'}) is None
    assert demo_db.lookup('hits', 'batters', {'year': 1999, 'name': '"Piazza"'}) is None
    assert demo_db.lookup('hits', 'batters', {'year': 1999, 'name': 5}) is None


def test_dict_lookup(demo_db):
    d = demo_db.dict_lookup('name', 'hits', 'batters', 'WHERE year == 1999')
    assert d['Olerud'] == 173
    assert d['Piazza'] == 162
    assert d['Alfonzo'] == 191


def test_dict_lookup_with_clause(demo_db):
    d = demo_db.dict_lookup('name', 'hits', 'batters', {'year': 1998})
    assert d['Olerud'] == 197
    assert d['Piazza'] == 137
    assert d['Alfonzo'] == 155


def test_unique_counts(demo_db):
    d = demo_db.unique_counts('batters', 'name')
    assert d['Olerud'] == 2
    assert d['Piazza'] == 3
    assert d['Alfonzo'] == 3


def test_sum(demo_db):
    assert demo_db.sum('batters', 'hits') == 1493
    assert demo_db.sum('batters', 'hits', {}) == 1493
    assert demo_db.sum('batters', 'hits', {'year': 2000}) == 478
    assert demo_db.sum('batters', 'hits', {'position': Position.FIRST_BASE}) == 516


def test_sum_counts(demo_db):
    d = demo_db.sum_counts('batters', 'hits', 'name')
    assert d['Olerud'] == 370
    assert d['Piazza'] == 455
    assert d['Alfonzo'] == 522


def test_insertion(demo_db):
    new_id = demo_db.insert('batters', {'name': 'Abayani', 'hits': 101, 'year': 2000, 'position': Position.LEFT_FIELD})
    assert new_id == 10
    assert demo_db.count('batters') == 10


def test_bulk_insert(demo_db):
    demo_db.bulk_insert('batters', ['name', 'year', 'hits', 'position'], [
        ('Ordóñez', 1999, 134, Position.SHORTSTOP),
        ('Ventrua', 1999, 177, Position.THIRD_BASE),
    ])

    assert demo_db.count('batters') == 11


def test_update(demo_db):
    # No match, just insert
    assert demo_db.count('batters') == 9
    row_id = demo_db.update('batters', {'name': 'McEwing', 'year': 2000, 'hits': 34}, ['name', 'year'])
    assert demo_db.count('batters') == 10
    assert row_id == 10

    # Actual update
    assert demo_db.lookup('hits', 'batters', {'name': 'Olerud', 'year': 1999}) == 173
    row_id = demo_db.update('batters', {'name': 'Olerud', 'year': 1999, 'hits': 334}, ['name', 'year'])
    assert demo_db.count('batters') == 10
    assert row_id == 4
    assert demo_db.lookup('hits', 'batters', {'name': 'Olerud', 'year': 1999}) == 334

    # Update by ID
    b_id = demo_db.lookup('id', 'batters', {'name': 'Zeile'})
    row_id = demo_db.update('batters', {'id': b_id, 'hits': 4})
    assert demo_db.count('batters') == 10
    assert row_id == b_id
    assert demo_db.lookup('hits', 'batters', {'name': 'Zeile'}) == 4


def test_none(demo_db):
    assert demo_db.count('batters', clause={'name': None}) == 0
    demo_db.insert('batters', {'year': 2002, 'hits': 5})
    clause = 'WHERE hits > 0 AND ' + demo_db.generate_clause({'name': None}, full=False)
    assert demo_db.count('batters', clause=clause) == 1


@pytest.fixture()
def demo_without_id_db():
    path = pathlib.Path('all_field.db')
    db = SQLiteDB(path, default_type='int')
    db.tables['batters'] = ['name', 'year', 'position']
    db.field_types['name'] = 'text'
    db.field_types['position'] = 'Position'
    db.register_custom_enum(Position)
    db.update_database_structure()
    for values in [
        ['Olerud', 1998, Position.FIRST_BASE],
        ['Piazza', 1998, Position.CATCHER],
        ['Alfonzo', 1998, Position.THIRD_BASE],
    ]:
        db.execute('INSERT INTO batters (name, year, position) VALUES(?, ?, ?)', values)

    yield db
    db.dispose()


def test_all_field_update(demo_without_id_db):
    db = demo_without_id_db

    # No match, just insert
    assert db.count('batters') == 3
    row_id = db.update('batters', {'name': 'McEwing', 'year': 2000}, ['name', 'year'])
    assert db.count('batters') == 4
    assert row_id is None

    # Actual update
    row_id = db.update('batters', {'name': 'Olerud', 'year': 1998}, ['name', 'year'])
    assert db.count('batters') == 4
    assert row_id is None


def test_accidental_open():
    db = SQLiteDB(pathlib.Path('empty.db'))
    db.update_database_structure()  # Update with no tables defined
    db.dispose()


@pytest.fixture()
def date_db():
    path = pathlib.Path('history.db')
    db = SQLiteDB(path)
    db.tables = {
        'great_moments': ['id', 'name', 'date'],
        'better_moments': ['id', 'name', 'datetime'],
    }
    db.field_types['id'] = 'int'
    db.field_types['date'] = 'date'
    db.field_types['datetime'] = 'timestamp'
    db.update_database_structure()

    yield db
    db.dispose()


def test_date_handling(date_db):
    # Test insertion
    date_db.execute('INSERT INTO great_moments (name, date) VALUES(?, ?)',
                    ['Declaration of Independence', datetime.date(1776, 7, 4)])

    assert date_db.count('great_moments') == 1

    # Check return types
    row = date_db.query_one('SELECT * FROM great_moments')
    assert row['id'] == 1
    assert row['name'] == 'Declaration of Independence'
    assert row['date'] == datetime.date(1776, 7, 4)
    assert isinstance(row['date'], datetime.date)

    # Insert two more
    date_db.insert('great_moments', {'name': 'Beethoven\'s Ninth Symphony', 'date': '1824-05-07'})
    date_db.update('great_moments', {'name': 'Rosalind Franklin born', 'date': datetime.date(1920, 7, 25)}, 'name')

    # Check output type
    assert isinstance(date_db.lookup('date', 'great_moments', {'name': "Beethoven's Ninth Symphony"}),
                      datetime.date)

    # Check clause generation
    event_name = date_db.lookup('name', 'great_moments', {'date': datetime.date(1920, 7, 25)})
    assert event_name == 'Rosalind Franklin born'


def test_datetime_handling(date_db):
    date_db.insert('better_moments', {'name': '2015 Game 1', 'datetime': datetime.datetime(2015, 10, 27, 20, 7)})
    date_db.insert('better_moments', {'name': '1986 Game 6', 'datetime': datetime.datetime(1986, 10, 25, 20, 30)})
    date_db.insert('better_moments', {'name': '1969 Game 5', 'datetime': datetime.datetime(1969, 10, 16)})

    # Check output type
    assert isinstance(date_db.lookup('datetime', 'better_moments', {'name': '2015 Game 1'}), datetime.datetime)

    # Check clause generation
    event_name = date_db.lookup('name', 'better_moments', {'datetime': datetime.datetime(1986, 10, 25, 20, 30)})
    assert event_name == '1986 Game 6'

    count = date_db.count('better_moments', clause='WHERE datetime > "1984-01-01"')
    assert count == 2



import sqlite3


class DatabaseError(sqlite3.Error):
    def __init__(self, error_s, command, parameters=None):
        s = f'{error_s}\nCommand: {command}'
        if parameters:
            s += f'\n\t{parameters}'
        sqlite3.Error.__init__(self, s)
        self.command = command
        self.parameters = parameters


class Row(sqlite3.Row):
    def __repr__(self):
        return str(dict(self))


PYTHON_SQL_TYPE_TRANSLATION = {
    'int': 'INTEGER',
    'float': 'REAL',
    'string': 'TEXT',
}


class SQLiteDB:
    """Core database structure that handles base sqlite3 interactions"""

    def __init__(self, database_path, default_type='TEXT', primary_keys=['id']):
        """
        Args:
            database_path (pathlib.Path): File to store the data
            default_type (str): The default SQL type to use
            primary_keys (list of strings): Fields that should automatically be marked as primary keys
        """
        self.raw_db = sqlite3.connect(str(database_path), detect_types=sqlite3.PARSE_DECLTYPES)
        self.path = database_path
        self.raw_db.row_factory = Row

        self.tables = {}
        self.field_types = {}
        self.default_type = default_type
        self.primary_keys = list(primary_keys)
        self.primary_key_per_table = {}
        self.adapters = {}
        self.converters = {}
        self.register_custom_type('bool', bool, int, lambda v: bool(v))
        self.q_strings = {}

    def register_custom_type(self, name, type_, adapter_fn, converter_fn):
        """Register a non-standard datatype.

        Args:
            name (str): The name of the custom type
            type_ (class): Python type
            adapter_fn (function): Translates the Python type to the sqlite3 type
            converter_fn (function): Translates the sqlite3 to the Python type
        """
        self.adapters[name] = adapter_fn
        self.converters[name] = converter_fn
        sqlite3.register_adapter(type_, adapter_fn)
        sqlite3.register_converter(name, converter_fn)

    def register_custom_enum(self, custom_enum_class):
        """Register an IntEnum

        Args:
            custom_enum_class (IntEnum): Class type of enum to register
        """
        self.register_custom_type(custom_enum_class.__name__,
                                  custom_enum_class,
                                  lambda d: d.value,
                                  lambda v: custom_enum_class(int(v)))

    def query_one(self, query):
        """Run the specified query and return the first result

        Args:
            query (str): SQL query to execute

        Returns:
            Row or None: The result of the query
        """
        try:
            cursor = self.raw_db.cursor()
            cursor.execute(query)
            return cursor.fetchone()
        except sqlite3.OperationalError as e:
            raise DatabaseError(str(e), query) from None

    def query(self, query):
        """Run the specified query and return the results

        Args:
            query (str): SQL query to execute

        Returns:
            Iterator(Row): The results of the query
        """
        try:
            cursor = self.raw_db.cursor()
            yield from cursor.execute(query)
        except (sqlite3.Error, ValueError) as e:
            raise DatabaseError(str(e), query) from None

    def execute(self, command, params=()):
        """Execute the given command with the parameters. Returns the cursor

        Args:
            command (str): SQL command to execute
            params (tuple): values to substitute into the placeholders

        Returns:
            Cursor: sqlite3 cursor for getting additional info like lastrowid
        """
        try:
            cur = self.raw_db.cursor()
            cur.execute(command, params)
            return cur
        except (sqlite3.Error, ValueError) as e:
            raise DatabaseError(str(e), command, params) from None

    def execute_many(self, command, objects):
        """Execute the given command multiple times.

        Args:
            command (str): SQL command to execute
            objects (list of tuples): The given command is run for each object/set of placeholder values
        """
        try:
            self.raw_db.executemany(command, objects)
        except (sqlite3.Error, ValueError) as e:
            raise DatabaseError(str(e), command, objects) from None

    def get_field_type(self, field, full=False):
        """Return a string representing the type of a given field.

        Args:
            field (str): The name of the field
            full (bool): If True, it returns other elements that would be used in the column definition
                         (i.e. PRIMARY KEY)
        Returns:
            str: SQL type for the given field
        """
        base = self.field_types.get(field, self.default_type)
        if base in self.converters:
            sql_type = base
        elif base in PYTHON_SQL_TYPE_TRANSLATION:
            sql_type = PYTHON_SQL_TYPE_TRANSLATION[base]
        else:
            sql_type = base.upper()

        if not full or field not in self.primary_keys:
            return sql_type
        else:
            return sql_type + ' PRIMARY KEY'

    def get_sql_table_types(self, table):
        """Create a dictionary mapping the name of each field in the table to its type in the actual db

        Args:
            table (str): Name of the table to get information about

        Returns:
            dict[str/str]: a mapping from field name to sql
        """
        type_map = {}
        for row in self.query(f'PRAGMA table_info("{table}")'):
            type_map[row['name']] = row['type']
        return type_map

    def update_database_structure(self):
        """Create or update the structure of all tables."""
        self.primary_key_per_table = {}
        for table, keys in self.tables.items():
            # Check if table exists
            table_exists = self.count('sqlite_master', f"WHERE type='table' AND name='{table}'") > 0
            if not table_exists:
                self.create_table(table, keys)
            else:
                self.update_table(table, keys)

            # Save primary key
            for key in keys:
                if key in self.primary_keys:
                    self.primary_key_per_table[table] = key

        if not self.tables:
            return

        # Cache strings consisting of a number of comma separated question marks
        for n in range(1, max(len(k) for k in self.tables.values()) + 1):
            self.q_strings[n] = ', '.join(['?'] * n)

    def create_table(self, table, keys):
        """Create a table with the given name and fields

        Args:
            table (str): Name of the table
            keys (str[]): Names of the fields
        """
        types = []
        for key in keys:
            tt = self.get_field_type(key, full=True)
            types.append(f'{key} {tt}')
        type_s = ', '.join(types)
        self.execute(f'CREATE TABLE {table} ({type_s})')

    def update_table(self, table, keys, field_mappings={}):
        """Update a table to have the given keys while preserving the data.

        Args:
            table (str): Name of the table
            keys (str[]): Keys that the table should have after this operation
            field_mappings (dict[str/str]): Mapping of new field names to old field names.
        """
        self.tables[table] = keys
        type_map = self.get_sql_table_types(table)

        fields_to_add = []
        old_fields = []
        new_fields = []
        needs_restructure = False

        for key in keys:
            if key in field_mappings:
                old_fields.append(field_mappings[key])
                new_fields.append(key)
                needs_restructure = True
            elif key in type_map:
                old_fields.append(key)
                new_fields.append(key)

                if type_map[key] != self.get_field_type(key):
                    needs_restructure = True
            else:
                fields_to_add.append(key)

        fields_to_remove = set(type_map.keys()) - set(keys)
        needs_restructure = needs_restructure or len(fields_to_remove)

        if not needs_restructure:
            # Can alter the table in-place
            for field in fields_to_add:
                tt = self.get_field_type(field, full=True)
                self.execute(f'ALTER TABLE {table} ADD COLUMN {field} {tt}')
            return

        temp_table_name = f'{table}_x'
        self.execute(f'ALTER TABLE {table} RENAME TO {temp_table_name}')
        self.create_table(table, self.tables[table])

        old_fields_s = ', '.join(old_fields)
        new_fields_s = ', '.join(new_fields)
        command = f'INSERT INTO {table}({new_fields_s}) SELECT {old_fields_s} FROM {temp_table_name}'
        self.execute(command)
        self.execute(f'DROP TABLE {temp_table_name}')

    def infer_database_structure(self):
        """Use the existing database entries to infer the tables and field_types"""
        for table in self.lookup_all('name', 'sqlite_master', "WHERE type='table'"):
            type_dict = self.get_sql_table_types(table)
            self.tables[table] = list(type_dict.keys())
            for field, type_name in type_dict.items():
                if type_name != self.default_type:
                    self.field_types[field] = type_name

    # Bonus "syntactic sugar" is provided in queries.py
    from ._queries import lookup_all, lookup, count, dict_lookup, unique_counts, sum_counts, insert, bulk_insert
    from ._queries import format_value, generate_clause, sum, update

    def reset(self, table=None):
        """Clear all or some of the data out of the database and recreate the table(s).

        Args:
            table (str/None): If specified, the name of the table to reset. Otherwise, all tables are reset.
        """
        db = self.raw_db.cursor()
        if table is None:
            tables = list(self.tables.keys())
        else:
            tables = [table]

        for table in tables:
            db.execute(f'DROP TABLE IF EXISTS {table}')

        self.update_database_structure()

    def write(self):
        """Commit the changes to the file."""
        self.raw_db.commit()

    def close(self, print_table_sizes=True):
        """Write data to database. Possibly print the number of rows in each table.

        Args:
            print_table_sizes (bool): Whether to print the table sizes
        """
        if print_table_sizes:
            print(self)
        self.write()
        self.raw_db.close()

    def dispose(self):
        """Used in tests to close the database and remove the file."""
        self.close(print_table_sizes=False)
        self.path.unlink()

    def __repr__(self):
        """String representing the number of rows in each table.

        Returns:
            str: the name and size of each table on its own row
        """
        s = ''
        for table in self.tables:
            s += f'{table}({self.count(table)})\n'
        return s



import pathlib
import yaml

from .sqlite_db import SQLiteDB


class MetroDB(SQLiteDB):
    """SQLiteDB that uses a yaml file to specify the database structure"""

    def __init__(self, key, folder=pathlib.Path('.'), extension='db', enums_to_register=[]):
        """Constructor

        Args:
            key (str): Name of the database. Used in constructing the filename
            folder (pathlib.Path): Folder for the database file, and maybe the yaml
            extension (str): The filename suffix for the database file
            enums_to_register (list): A list of enums to register
        """
        SQLiteDB.__init__(self, folder / f'{key}.{extension}')
        self.folder = folder
        self.key = key

        for custom_enum_class in enums_to_register:
            self.register_custom_enum(custom_enum_class)

    def load_yaml(self, structure_filepath=None, structure_key=None):
        """Manually load the yaml file.

        By default, if the database file is path/name.db, the yaml file is path/name.yaml

        However, you can also specify the full path with structure_filepath or change the name
        by specifying structure_key (so that the yaml will instead be path/structure_key.yaml)

        Args:
            structure_filepath (pathlib.Path): Optional full path to the yaml file
            structure_key (str): Optional stem for the yaml file
        """
        if structure_filepath is None:
            if structure_key is None:
                structure_key = self.key

            structure_filepath = self.folder / f'{structure_key}.yaml'
        elif structure_key:
            raise RuntimeError('Cannot specify structure_filepath AND structure_key')

        db_structure = yaml.safe_load(open(structure_filepath))
        self.tables = db_structure['tables']
        self.field_types = db_structure.get('types', db_structure.get('field_types', {}))
        self.default_type = db_structure.get('default_type', self.default_type)
        self.primary_keys = db_structure.get('primary_keys', self.primary_keys)

    def update_database_structure(self):
        """Create or update the structure of all tables.

        Loads the yaml automatically if it has not been done already."""
        if not self.tables:
            self.load_yaml()
        SQLiteDB.update_database_structure(self)

    def __enter__(self):
        self.update_database_structure()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()



import argparse
from tabulate import tabulate, multiline_formats

from . import SQLiteDB

multiline_formats['fancy_outline'] = 'fancy_outline'  # Hack to fix tabulate bug


def main(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('db_path')
    parser.add_argument('n', nargs='?', default=10, type=int,
                        help='Number of rows of each table to display. -1 for all')
    parser.add_argument('-s', '--style', choices=['simple', 'grid', 'plain', 'fancy_outline'], default='fancy_outline')
    args = parser.parse_args(argv)

    db = SQLiteDB(args.db_path)
    db.infer_database_structure()

    for table in db.lookup_all('name', 'sqlite_master', 'WHERE type="table"'):
        query = f'SELECT * FROM {table}'
        if args.n >= 0:
            query += f' LIMIT {args.n}'

        results = list(db.query(query))
        headers = []
        for key in db.tables[table]:
            type_name = db.get_field_type(key)
            headers.append(f'{key}\n({type_name})')
        count = db.count(table)
        if args.n >= 0 and count > args.n:
            extra_row = [f'...{count - args.n} more...']
            while len(extra_row) < len(headers):
                extra_row.append('...')
            results.append(extra_row)
        output = tabulate(results, headers=headers, tablefmt=args.style)
        length = max(len(s) for s in output.split('\n'))
        print(('{:^' + str(length) + '}').format(table))
        print(output)
        print()



import pathlib
import pytest
from metro_db import SQLiteDB, DatabaseError
from enum import IntEnum


@pytest.fixture()
def basic_db():
    path = pathlib.Path('basic.db')
    db = SQLiteDB(path)
    db.tables['people'] = ['name', 'age', 'grade', 'present']
    db.field_types['age'] = 'int'
    db.field_types['grade'] = 'float'
    db.field_types['present'] = 'bool'
    yield db
    db.dispose()


def test_creation(basic_db):
    basic_db.update_database_structure()


def test_insertion(basic_db):
    basic_db.update_database_structure()
    basic_db.execute('INSERT INTO people (name, age, grade, present) VALUES(?, ?, ?, ?)', [1, 1, 1, 1])

    assert basic_db.count('people') == 1

    # Check return types
    row = basic_db.query_one('SELECT * FROM people')
    assert row['name'] == '1'
    assert row['age'] == 1
    assert row['grade'] == 1    # 1 == 1.0
    assert isinstance(row['grade'], float)
    assert row['present'] is True

    # Check string representation of row
    assert str(row) == "{'name': '1', 'age': 1, 'grade': 1.0, 'present': True}"

    # Check repr
    assert str(basic_db) == 'people(1)\n'


def test_insertion_problem(basic_db):
    basic_db.update_database_structure()
    command = 'INSERT INTO people (name, age, grade, present) VALUES(?, ?, ?)'  # Missing ?
    values = [1, 1, 1, 1]
    with pytest.raises(DatabaseError) as e:
        basic_db.execute(command, values)

    assert '3 values for 4 columns' in str(e.value)
    assert command == e.value.command
    assert command in str(e.value)
    assert values == e.value.parameters


def test_empty_query(basic_db):
    basic_db.update_database_structure()
    assert basic_db.count('people') == 0
    row = basic_db.query_one('SELECT * FROM people')
    assert row is None


def test_query_problems(basic_db):
    basic_db.update_database_structure()
    command = 'SELECT FROM PEOPLE'  # missing fields
    with pytest.raises(DatabaseError) as e:
        basic_db.query_one(command)

    assert 'syntax error' in str(e.value)
    assert command == e.value.command
    assert command in str(e.value)
    assert e.value.parameters is None

    with pytest.raises(DatabaseError) as e:
        list(basic_db.query(command))

    assert 'syntax error' in str(e.value)
    assert command == e.value.command
    assert command in str(e.value)
    assert e.value.parameters is None


def test_execute_many(basic_db):
    basic_db.update_database_structure()
    basic_db.execute_many('INSERT INTO people (name, age, grade, present) VALUES(?, ?, ?, ?)',
                          [['David', 25, 98.6, True],
                           ['Elise', 24, 99.1, True]])
    assert basic_db.count('people') == 2

    # Test reset
    basic_db.reset()
    assert basic_db.count('people') == 0


def test_execute_many_problems(basic_db):
    basic_db.update_database_structure()
    with pytest.raises(DatabaseError) as e:
        # missing table name
        basic_db.execute_many('INSERT INTO  (name, age, grade, present) VALUES(?, ?, ?, ?)',
                              [['David', 25, 98.6, True],
                               ['Elise', 24, 99.1, True]])
    assert 'syntax error' in str(e.value)


def test_inference(basic_db):
    basic_db.update_database_structure()

    # Open second object to same path
    db = SQLiteDB(pathlib.Path('basic.db'))
    db.infer_database_structure()
    assert db.tables == {'people': ['name', 'age', 'grade', 'present']}
    db.close(print_table_sizes=False)


def test_column_add(basic_db):
    basic_db.update_database_structure()
    basic_db.execute('INSERT INTO people (name, age, grade, present) VALUES(?, ?, ?, ?)', [1, 1, 1, 1])

    row = basic_db.query_one('SELECT * FROM people')
    assert len(row) == 4

    basic_db.tables['people'].append('new_text_field')
    basic_db.update_database_structure()

    row = basic_db.query_one('SELECT * FROM people')
    assert len(row) == 5


def test_column_removal(basic_db):
    basic_db.update_database_structure()
    basic_db.execute('INSERT INTO people (name, age, grade, present) VALUES(?, ?, ?, ?)', [1, 1, 1, 1])

    row = basic_db.query_one('SELECT * FROM people')
    assert len(row) == 4

    basic_db.tables['people'].pop(-1)  # Remove present
    basic_db.update_database_structure()

    row = basic_db.query_one('SELECT * FROM people')
    assert len(row) == 3


def test_quiet_close(capsys):
    path = pathlib.Path('basic.db')
    db = SQLiteDB(path)
    db.tables['people'] = ['name', 'age', 'grade', 'present']
    db.update_database_structure()
    db.dispose()
    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == ''


def test_close(capsys):
    path = pathlib.Path('basic.db')
    db = SQLiteDB(path)
    db.tables['people'] = ['name', 'age', 'grade', 'present']
    db.update_database_structure()
    db.close(print_table_sizes=True)
    path.unlink()
    captured = capsys.readouterr()
    assert captured.out == 'people(0)\n\n'
    assert captured.err == ''


def test_two_tables(basic_db):
    basic_db.tables['cars'] = ['make', 'model']
    basic_db.update_database_structure()
    basic_db.execute_many('INSERT INTO people (name, age, grade, present) VALUES(?, ?, ?, ?)',
                          [['David', 25, 98.6, True],
                           ['Elise', 24, 99.1, True]])
    basic_db.execute('INSERT INTO cars (make, model) VALUES(?, ?)', ('Toyota', 'Prius'))
    assert basic_db.count('people') == 2
    assert basic_db.count('cars') == 1

    # Test reset
    basic_db.reset('people')
    assert basic_db.count('people') == 0
    assert basic_db.count('cars') == 1

    basic_db.reset()
    assert basic_db.count('people') == 0
    assert basic_db.count('cars') == 0


def test_enum(basic_db):
    class Status(IntEnum):
        HERE = 1
        ABSENT = 2

    basic_db.register_custom_enum(Status)
    basic_db.field_types['present'] = 'Status'
    basic_db.update_database_structure()

    basic_db.execute('INSERT INTO people (name, present) VALUES(?, ?)', ['Angelica', None])
    basic_db.execute('INSERT INTO people (name, present) VALUES(?, ?)', ['Eliza', Status.HERE])
    basic_db.execute('INSERT INTO people (name, present) VALUES(?, ?)', ['Peggy', Status.ABSENT])

    assert basic_db.query_one('SELECT present FROM people WHERE name="Eliza"')[0] == Status.HERE
    assert basic_db.query_one('SELECT present FROM people WHERE name="Peggy"')[0] == Status.ABSENT
    assert basic_db.query_one('SELECT present FROM people WHERE name="Angelica"')[0] is None


def test_column_alter(basic_db):
    basic_db.update_database_structure()
    basic_db.execute_many('INSERT INTO people (name, age, grade, present) VALUES(?, ?, ?, ?)',
                          [['David', 25, 98.6, True],
                           ['Elise', 24, 99.1, True]])

    basic_db.field_types['temperature'] = 'float'
    basic_db.update_table('people', ['name', 'age', 'temperature', 'present'], {'temperature': 'grade'})

    assert basic_db.lookup('temperature', 'people', 'WHERE name="David"') == 98.6

    basic_db.field_types['present'] = 'int'
    basic_db.update_database_structure()
    assert basic_db.lookup('present', 'people', 'WHERE name="David"') == 1



def format_value(self, field, value):
    """If the field's type is text, surround with quotes.

    Args:
        field (str): Name of the field
        value: The value to format

    Returns:
        str: String to be inserted into an SQL query.
    """
    ft = self.get_field_type(field)
    if ft == 'TEXT':
        if not isinstance(value, str):
            value = str(value)
        if '"' in value:
            if "'" in value:
                return '"{}"'.format(value.replace('"', '""'))
            else:
                return f"'{value}'"
        else:
            return f'"{value}"'
    elif ft in self.adapters:
        return self.adapters[ft](value)
    elif ft in ['DATE', 'TIMESTAMP']:
        return f'"{value}"'
    else:
        return str(value)


def generate_clause(self, value_dict, operator='AND', full=True):
    """Generate a string clause that matches the value dictionary. If full, include the keyword WHERE

    Args:
        value_dict (dict): Values that should match. Keys are string fieldnames.
        operator (str): Operator to link subclauses with. Default is AND.
        full (bool): If true, includes the WHERE string at the beginning.

    Returns:
        str
    """
    if not value_dict:
        return ''
    pieces = []
    for key, value in value_dict.items():
        if value is None:
            pieces.append(f'{key} IS NULL')
        else:
            pieces.append('{}={}'.format(key, self.format_value(key, value)))

    clause = f' {operator} '.join(pieces)
    if full:
        return f'WHERE {clause}'
    else:
        return clause


def lookup_all(self, field, table, clause='', distinct=False):
    """Run a SELECT command with the specified field, table and clause, return the matching values.

    Args:
        field (str): Name of field to return
        table (str): Name of table to query
        clause (str/dict): Optional clause to add to end of query. If dict, use generate_clause to translate to str.
        distinct (bool): If true, add DISTINCT keyword before field name

    Returns:
        iterator: All the values that match the query
    """
    field_s = field if not distinct else f'DISTINCT {field}'
    if not isinstance(clause, str):
        clause = self.generate_clause(clause)
    for row in self.query(f'SELECT {field_s} FROM {table} {clause}'):
        yield row[0]


def lookup(self, field, table, clause=''):
    """Run a SELECT command and return the first (only?) value.

    Args:
        field (str): Name of field to return
        table (str): Name of table to query
        clause (str/dict): Optional clause to add to end of query. If dict, use generate_clause to translate to str.

    Returns:
        The value or None"""
    if not isinstance(clause, str):
        clause = self.generate_clause(clause)
    result = self.query_one(f'SELECT {field} FROM {table} {clause}')
    if result:
        return result[0]


def count(self, table, clause=''):
    """Return the number of results for a given query.

    Args:
        table (str): Name of table to query
        clause (str/dict): Optional clause to add to query. If dict, use generate_clause to translate to str.

    Returns:
        int: Number of rows in the table (that match the clause)
    """
    return self.lookup('COUNT(*)', table, clause)


def dict_lookup(self, key_field, value_field, table, clause=''):
    """Return a dictionary mapping the key_field to the value_field for some query.

    Args:
        key_field (str): The name of the field that should be the key in the dictionary
        value_field (str): The name of the field that should be the value in the dictionary
        table (str): The name of the table to query
        clause (str/dict): Optional clause to add to query. If dict, use generate_clause to translate to str.

    Returns:
        dict: All of the rows returned by the query formatted into a dictionary

    """
    if not isinstance(clause, str):
        clause = self.generate_clause(clause)
    results = self.query(f'SELECT {key_field}, {value_field} FROM {table} {clause}')
    return {d[key_field]: d[value_field] for d in results}


def unique_counts(self, table, ident_field):
    """Return a dictionary mapping the different values of the ident_field column to how many times each appears.

    Args:
        table (str): The name of the table to query
        ident_field (str): The name of the field to count

    Returns:
        dict: All of the unique values of ident_field mapped to an integer count
    """
    return self.dict_lookup(ident_field, 'COUNT(*)', table, 'GROUP BY ' + ident_field)


def sum(self, table, value_field, clause=''):
    """Return the sum of the value_field column.

    Args:
        table (str): The name of the table to query
        value_field (str): The name of the field (presumed to have integer type)
        clause (str/dict): Optional clause to add to query. If dict, use generate_clause to translate to str.

    Returns:
        int: The sum of all the values of value_field matching the query.
    """
    return self.lookup(f'SUM({value_field})', table, clause)


def sum_counts(self, table, value_field, ident_field):
    """Return the values of the ident_field column mapped to the sum of the value_field column.

    Args:
        table (str): The name of the table to query
        value_field (str): The name of the value field (presumed to have an integer type)
        ident_field (str): The name of the identifier field, used as the key

    Returns:
        dict: A mapping from all the unique values of ident_field to the sum of all the matching value_field entries
    """
    return self.dict_lookup(ident_field, f'SUM({value_field})', table, 'GROUP BY ' + ident_field)


def insert(self, table, row_dict):
    """Insert the given row into the table.

    Args:
        table (str): The name of the table to insert into
        row_dict (dict): A dictionary where the keys are string field names, and the values are, well, the values

    Returns:
        int: the lastrowid (i.e. probably the primary key of the inserted row)
    """
    keys = row_dict.keys()

    values = [row_dict.get(k) for k in keys]
    key_s = ', '.join(keys)
    n = len(values)

    cur = self.execute(f'INSERT INTO {table} ({key_s}) VALUES({self.q_strings[n]})', values)
    return cur.lastrowid


def bulk_insert(self, table, fields, rows):
    """Insert multiple rows into the table at a time

    Args:
        table (str): The name of the table to insert into
        fields (str[]): The names of the fields
        rows (tuple[]): Each tuple is the values for the fields.
                        The length of each tuple should match the length of fields
    """
    n = len(fields)
    key_s = ', '.join(fields)
    self.execute_many(f'INSERT INTO {table} ({key_s}) VALUES({self.q_strings[n]})', rows)


def update(self, table, row_dict, replace_key='id'):
    """If there's a row where the key value matches the row_dict's value, update it. Otherwise, insert it.

    Args:
        table (str): The name of the table
        row_dict (dict): A dictionary where the keys are string field names, and the values are, well, the values
        replace_key (str/list): If replace_key is a string, then the row where the field matches is updated.
                                Otherwise, it specifies multiple field names that have to match.

    Returns:
        int: The row id of the new or old row (emulating lastrowid)
    """
    if isinstance(replace_key, str):
        clause = self.generate_clause({replace_key: row_dict[replace_key]})
    else:
        clause = self.generate_clause({key: row_dict[key] for key in replace_key})

    existing = self.query_one(f'SELECT * FROM {table} {clause}')
    if not existing:
        # If no matches, just insert
        self.insert(table, row_dict)
        existing = self.query_one(f'SELECT * FROM {table} {clause}')
    else:
        field_qs = []
        values = []
        for k in row_dict.keys():
            if isinstance(replace_key, str) and k == replace_key:
                continue
            elif isinstance(replace_key, (list, dict)) and k in replace_key:
                continue
            values.append(row_dict[k])
            field_qs.append(f'{k}=?')

        if field_qs:
            field_s = ', '.join(field_qs)
            query = f'UPDATE {table} SET {field_s} ' + clause
            self.execute(query, values)

    # Determine proper return
    if table not in self.primary_key_per_table:
        # No primary key, return nothing
        return
    return_key = self.primary_key_per_table[table]

    if return_key in row_dict:
        return row_dict[return_key]
    else:
        return existing[return_key]



import pathlib
import pytest
from metro_db import SQLiteDB
from metro_db.peek import main as peek


@pytest.fixture()
def basic_db():
    path = pathlib.Path('basic.db')
    db = SQLiteDB(path)
    db.tables['people'] = ['name', 'age', 'grade', 'present']
    db.field_types['age'] = 'int'
    db.field_types['grade'] = 'float'
    db.field_types['present'] = 'bool'
    db.update_database_structure()

    db.execute_many('INSERT INTO people (name, age, grade, present) VALUES(?, ?, ?, ?)',
                    [['David', 25, 98.6, True],
                     ['Elise', 24, 99.1, True]])
    db.write()
    yield db
    db.dispose()


def test_peek_basic(basic_db, capsys):
    peek(['basic.db'])
    captured = capsys.readouterr()
    assert captured.out == open('tests/out/basic_out.txt').read()
    assert captured.err == ''


def test_peek_n1(basic_db, capsys):
    peek(['basic.db', '1'])
    captured = capsys.readouterr()
    assert captured.out == open('tests/out/basic_n1.txt').read()
    assert captured.err == ''


def test_peek_n10(basic_db, capsys):
    peek(['basic.db', '10'])
    captured = capsys.readouterr()
    assert captured.out == open('tests/out/basic_out.txt').read()
    assert captured.err == ''



import pathlib
import pytest
from metro_db import MetroDB
from enum import IntEnum


class Role(IntEnum):
    PROTAGONIST = 1
    LOVE_INTEREST = 2


TEST_FOLDER = pathlib.Path('tests')


def test_metro_basic():
    db = MetroDB('metro', folder=TEST_FOLDER, enums_to_register=[Role])
    db.update_database_structure()
    assert len(db.tables) == 1
    assert 'characters' in db.tables
    db.dispose()


def test_structure_filepath():
    db = MetroDB('other', folder=TEST_FOLDER)
    db.load_yaml(db.folder / 'metro.yaml')
    assert len(db.tables) == 1
    assert 'characters' in db.tables
    db.dispose()


def test_structure_key():
    db = MetroDB('other', folder=TEST_FOLDER)
    db.load_yaml(structure_key='metro')
    assert len(db.tables) == 1
    assert 'characters' in db.tables
    db.dispose()


def test_invalid_spec():
    db = MetroDB('other', folder=TEST_FOLDER)
    with pytest.raises(RuntimeError):
        db.load_yaml(db.folder / 'metro.yaml', 'x')
    db.dispose()


def test_scope():
    with MetroDB('metro', folder=TEST_FOLDER, enums_to_register=[Role]) as db:
        assert len(db.tables) == 1
        assert 'characters' in db.tables

    pathlib.Path('tests/metro.db').unlink()



from setuptools import setup, find_packages

extra_test = [
    'pytest>=4',
    'pytest-cov>=2',
]

setup(
    name='metro_db',
    description='A wrapper around the sqlite3 database for easy database development',
    version='v0.2.0',
    url='https://github.com/DLu/metro_db',
    author='David V. Lu!!',
    author_email='davidvlu@gmail.com',
    packages=find_packages(),
    install_requires=[
        'pyyaml',
        'tabulate',
    ],
    entry_points={
        'console_scripts': [
            'metro_db=metro_db.peek:main',
        ],
    },
    extras_require={
        'dev': extra_test,
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Database',
    ],
)
